/*
BSD 2-Clause License

Copyright (c) 2011 - 2021, Franticware
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <FL/Fl.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Slider.H>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Value_Input.H>
#include <FL/Fl_Native_File_Chooser.H>
#include <FL/fl_draw.H>
#include <FL/fl_message.H>
#include <vector>
#include <string>
#include <cmath>

#define INITIAL_SIZE 2
#define INITIAL_ZOOM -14
#define INITIAL_COUNT 15
#define INITIAL_ANGLE 0

inline double myfrac(double d) // calculate positive fraction
{
    return d - std::floor(d);
}

// rainbow color computation
void rainbow(unsigned char& r, unsigned char& g, unsigned char& b, double x, double lighten, double darken)
{
    double rf = std::fabs(x*6-3)-1; if (rf < 0) rf = 0; else if (rf > 1) rf = 1;
    double gf = -std::fabs(x*6-2)+2; if (gf < 0) gf = 0; else if (gf > 1) gf = 1;
    double bf = -std::fabs(x*6-4)+2; if (bf < 0) bf = 0; else if (bf > 1) bf = 1;

    double rfc = 1.0-rf;
    double gfc = 1.0-gf;
    double bfc = 1.0-bf;

    r = 255.0*(rf+rfc*lighten)*darken;
    g = 255.0*(gf+gfc*lighten)*darken;
    b = 255.0*(bf+bfc*lighten)*darken;
}

// voronoi point
struct Vorop
{
    double x, y;
    unsigned char r, g, b;
    void generatePoint(int colors_sel)
    {
        x = (rand()%1000)*0.001;
        y = (rand()%1000)*0.001;

        if (x+y>1)
        {
            x = 1-x;
            y = 1-y;
        }
        switch (colors_sel) // color generator selection
        {
            case 0:
                rainbow(r, g, b, (rand()%10000)*0.0001, 0.0, 1.0);
                break;
            case 1:
                rainbow(r, g, b, (rand()%10000)*0.0001, (rand()%10001)*0.0001, 1.0);
                break;
            case 2:
                rainbow(r, g, b, (rand()%10000)*0.0001, 0.0, (rand()%10001)*0.0001);
                break;
            case 3:
                r=g=b = rand();
                break;
            case 4:
                r = rand();
                g = rand();
                b = rand();
                break;
        }
    }
};

// kaleidoscope
struct Voros {
    Voros()
    {
        m_size = INITIAL_SIZE;
        m_zoom = INITIAL_ZOOM;
        m_angle = INITIAL_ANGLE;
        m_colors_sel = 0;
        m_points.resize(INITIAL_COUNT);
        for (unsigned int i = 0; i != m_points.size(); ++i)
        {
            m_points[i].generatePoint(m_colors_sel);
        }
        m_pixels.resize(m_size*m_size*3);
        recompute();
    }
    void resize(unsigned int newsize) // change image size
    {
        if (m_size != newsize) {
            m_size = newsize;
            m_pixels.resize(m_size*m_size*3);
            recompute();
        }
    }
    void rotate(int newangle) // change angle
    {
        int anglediff = m_angle - newangle;
        if (anglediff < 0) anglediff = -anglediff;
        // rearrange and recolor a few beads
        int numchanges = m_points.size()/15;
        if (numchanges == 0) numchanges = 1;
        numchanges *= anglediff;
        for (int i = 0; i < numchanges; ++i) {
            m_points[rand()%m_points.size()].generatePoint(m_colors_sel);
        }
        m_angle = newangle;
        recompute();
    }
    void count(int countPrm) // change number of beads
    {
        int countPrev = m_points.size();
        m_points.resize(countPrm);
        if (countPrev < countPrm) {
            for (int i = countPrev; i != countPrm; ++i)
                m_points[i].generatePoint(m_colors_sel);
        }
        recompute();
    }
    void zoom(int zoomPrm) // change zoom
    {
        m_zoom = zoomPrm;
        recompute();
    }
    double computeZoom(int zoom) // zoom step to actual zoom value
    {
        return std::exp(double(zoom)*0.1);
    }
    void recompute() // calculate pixel colors
    {
        double deg2rad = M_PI/180.0;
        double l = computeZoom(m_zoom);
        double a0 = cos((-m_angle*4)*deg2rad)*l; // points of the base triangle
        double a1 = sin((-m_angle*4)*deg2rad)*l;
        double b0 = cos((-m_angle*4+120)*deg2rad)*l;
        double b1 = sin((-m_angle*4+120)*deg2rad)*l;
        double c0 = cos((-m_angle*4+240)*deg2rad)*l;
        double c1 = sin((-m_angle*4+240)*deg2rad)*l;
        double u0 = b0-a0; // first pair of vectors
        double u1 = b1-a1;
        double v0 = c0-a0;
        double v1 = c1-a1;

        double uu0 = c0-b0; // second pair of vectors
        double uu1 = c1-b1;
        double vv0 = a0-b0;
        double vv1 = a1-b1;

        for (unsigned int y = 0; y != m_size; ++y)
        {
            for (unsigned int x = 0; x != m_size; ++x)
            {
                double d0 = 2.0*double(x)/double(m_size)-1.0; // graphics area coords are in range <-1; 1)
                double d1 = 2.0*double(y)/double(m_size)-1.0;

                double r = (v0*d1-v1*d0+v1*a0-v0*a1)/(v0*u1-v1*u0); // calculate the parameter of sheared 2D space
                double s = (u0*d1-u1*d0+u1*a0-u0*a1)/(u0*v1-u1*v0);

                double r1 = (vv0*d1-vv1*d0+vv1*b0-vv0*b1)/(vv0*uu1-vv1*uu0);
                double s1 = (uu0*d1-uu1*d0+uu1*b0-uu0*b1)/(uu0*vv1-uu1*vv0);

                double re=myfrac(r); // parameter in the rhombus
                double se=myfrac(s);

                double re1=myfrac(r1);
                double se1=myfrac(s1);

                int ri = r+60000; // global rhombus placement (shifted to be always positive)
                int si = s+30000;

                double rf = 0; // final coordinates in the triangle (respecting mirroring/symmetry)
                double sf = 0;

                switch ((ri-si)%3) { // mirorring/symmetry decision
                case 0:
                    if (re+se > 1)
                    { // decision, which half of the rhombus
                        rf = 1.0-se;
                        sf = 1.0-re;
                    }
                    else
                    {
                        rf = re;
                        sf = se;
                    }
                    break;
                case 1:
                    if (re+se > 1)
                    {
                        rf = 1.0-se1;
                        sf = 1.0-re1;
                    }
                    else
                    {
                        rf = se1;
                        sf = re;
                    }
                    break;
                case 2:
                    if (re+se > 1)
                    {
                        rf = 1.0-re;
                        sf = 1.0-se1;
                    }
                    else
                    {
                        rf = re1;
                        sf = se1;
                    }
                    break;
                }

                unsigned char red = 0;//rf*255;(xf+1)*127;
                unsigned char green = 0;//sf*255;(yf+1)*127;
                unsigned char blue = 0;
                double dist2 = 0;
                // calculation of voronoi points
                for (int i = 0; i != int(m_points.size()); ++i)
                {
                    double newdist2 =
                        (rf-m_points[i].x)*(rf-m_points[i].x)+
                        (sf-m_points[i].y)*(sf-m_points[i].y);
                    if (newdist2 < dist2 || i == 0)
                    {
                        dist2 = newdist2;
                        red = m_points[i].r;
                        green = m_points[i].g;
                        blue = m_points[i].b;
                    }
                }
                m_pixels[(x+y*m_size)*3+0] = red;
                m_pixels[(x+y*m_size)*3+1] = green;
                m_pixels[(x+y*m_size)*3+2] = blue;
            }
        }
    }
    void save_tga(const char* filename) // save the image
    {
        FILE* fin = fopen(filename, "wb");
        if (fin != NULL)
        {
            unsigned char tgaheader[] = {
                0, //0 	1 byte 	IDLength
                0, //1 	1 byte 	ColorMapType
                2, //2 	1 byte 	ImageType
                0,0, //3 	2 byte 	CMapStart
                0,0, //5 	2 byte 	CMapLength
                0, //7 	1 byte 	CMapDepth
                0,0, //8 	2 byte 	XOffset
                0,0 //10 	2 byte 	YOffset
                };
            unsigned short tgaheader2[] = {
                (unsigned short)m_size,//12 	2 byte 	Width
                (unsigned short)m_size//14 	2 byte 	Height
                };
            unsigned char tgaheader3[] = {
                24, //16 	1 byte 	PixelDepth
                32 //17 	1 byte 	ImageDescriptor
                };
            fwrite(tgaheader, 1, sizeof(tgaheader), fin);
            fwrite(tgaheader2, 1, sizeof(tgaheader2), fin);
            fwrite(tgaheader3, 1, sizeof(tgaheader3), fin);
            for (unsigned int i = 0; i != m_size*m_size; ++i)
            {
                fwrite(&m_pixels[i*3+2], 1, 1, fin);
                fwrite(&m_pixels[i*3+1], 1, 1, fin);
                fwrite(&m_pixels[i*3+0], 1, 1, fin);
            }
            fclose(fin);
        }
    }
    void open(const char* filename)
    {
        m_zoom = INITIAL_ZOOM;
        m_angle = INITIAL_ANGLE;
        m_points.clear();
        int i = 0;
        FILE* fin = fopen(filename, "rb");
        if (fin != NULL)
        {
            if (fread(&i, sizeof(int), 1, fin) == 1)
                m_angle = i;
            if (fread(&i, sizeof(int), 1, fin) == 1)
                m_zoom = i;
            double voropd_array[2];
            unsigned char voropb_array[3];
            while (fread(voropd_array, sizeof(voropd_array), 1, fin) == 1 && fread(voropb_array, sizeof(voropb_array), 1, fin) == 1)
            {
                Vorop vorop = {
                    voropd_array[0],
                    voropd_array[1],
                    voropb_array[0],
                    voropb_array[1],
                    voropb_array[2]
                    };
                m_points.push_back(vorop);
            }
            fclose(fin);
        }
        if (m_points.empty())
        {
            Vorop vorop = {0, 0, 0, 0, 0};
            m_points.push_back(vorop);
        }
        recompute();
    }
    void save(const char* filename)
    {
        FILE* fout = fopen(filename, "wb");
        if (fout != NULL)
        {
            fwrite(&m_angle, sizeof(int), 1, fout);
            fwrite(&m_zoom, sizeof(int), 1, fout);
            for (unsigned int i = 0; i != m_points.size(); ++i)
            {
                fwrite(&m_points[i].x, sizeof(double), 1, fout);
                fwrite(&m_points[i].y, sizeof(double), 1, fout);
                fwrite(&m_points[i].r, 1, 1, fout);
                fwrite(&m_points[i].g, 1, 1, fout);
                fwrite(&m_points[i].b, 1, 1, fout);
            }
            fclose(fout);
        }
    }
    std::vector<unsigned char> m_pixels; //
    std::vector<Vorop> m_points; // voronoi points
    unsigned int m_size; //
    int m_angle; //
    int m_zoom; //
    int m_colors_sel; // selected color generator
};

class Drawing : public Fl_Widget
{
    Voros m_voros;

    void draw()
    {
        fl_push_clip(x(),y(),w(),h());
        fl_color(FL_BACKGROUND_COLOR);
        fl_rectf(x(),y(),w(),h());

        int width = w();
        int height = h();
        int xstart;
        if (width > height) xstart = (width-height)>>1;
        else xstart = 0;
        m_voros.resize(height);
        fl_draw_image(&(m_voros.m_pixels[0]), xstart+x(), y(), height, height, 3, 0);
        fl_pop_clip();
    }

public:
    Drawing(int X,int Y,int W,int H) : Fl_Widget(X,Y,W,H) {}
    void angle(int anglePrm) // change angle
    {
        m_voros.rotate(anglePrm);
        redraw();
    }
    void count(int countPrm) // change bead count
    {
        m_voros.count(countPrm);
        redraw();
    }
    void zoom(int zoomPrm) // change zoom
    {
        m_voros.zoom(zoomPrm);
        redraw();
    }
    void colors(int colorsPrm) // change color generator
    {
        m_voros.m_colors_sel = colorsPrm;
    }
    void save_tga(const char* filename, int size)
    {
        Voros tmp(m_voros);
        tmp.resize(size);
        tmp.save_tga(filename);
    }
    void open(const char* filename)
    {
        m_voros.open(filename);
        redraw();
        m_sliders[0]->value(m_voros.m_angle);
        m_sliders[1]->value(m_voros.m_zoom);
        m_sliders[2]->value(m_voros.m_points.size());
    }
    void save(const char* filename)
    {
        m_voros.save(filename);
    }
    Fl_Slider* m_sliders[3];
};

struct ColorsSelData
{
    Drawing* drawing;
    int val;
};

struct ExportData
{
    Drawing* drawing;
    Fl_Value_Input* size;
};

struct OpenSaveData
{
    OpenSaveData(Drawing* drawing_prm)
    {
        drawing = drawing_prm;
        isfilename = false;
    }
    Drawing* drawing;
    std::string filename;
    bool isfilename;
};

static void slAngle_cb(Fl_Widget* widget, void* data)
{
    int val = int(((Fl_Slider*)widget)->value());
    ((Drawing*)data)->angle(val);
}

static void slZoom_cb(Fl_Widget* widget, void* data)
{
    int val = int(((Fl_Slider*)widget)->value());
    ((Drawing*)data)->zoom(val);
}

static void slCount_cb(Fl_Widget* widget, void* data)
{
    int val = int(((Fl_Slider*)widget)->value());
    ((Drawing*)data)->count(val);
}

static void chColors_cb(Fl_Widget* /*widget*/, void* data)
{
    int val = ((ColorsSelData*)data)->val;
    ((ColorsSelData*)data)->drawing->colors(val);
}

bool endswith(const char* str, const char* end)
{
    if (strlen(str) < strlen(end))
        return false;
    for (unsigned int i = 0; i != strlen(end); ++i)
    {
        if (str[strlen(str)-1-i] != end[strlen(end)-1-i])
            return false;
    }
    return true;
}

static void miExport_cb(Fl_Widget* /*widget*/, void* data)
{
    Fl_Native_File_Chooser fc(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
    fc.filter("TGA Images\t*.{tga,TGA}\nAll Files\t*");
    switch (fc.show())
    {
    case -1:
        fl_message("%s", fc.errmsg());
        break;
    case 1:
        break;
    default:
        {
            int size = ((ExportData*)data)->size->value();
            if (size < 1)
                size = 1;
            std::string tgafilename = fc.filename();
            if (!endswith(tgafilename.c_str(), ".tga") && !endswith(tgafilename.c_str(), ".TGA"))
                tgafilename += ".tga";
            ((ExportData*)data)->drawing->save_tga(tgafilename.c_str(), size);
        }
        break;
    }
}

static void miOpen_cb(Fl_Widget* /*widget*/, void* data)
{
    Fl_Native_File_Chooser fc(Fl_Native_File_Chooser::BROWSE_FILE);
    fc.filter("Kaleidoscope files\t*.kaleido\nAll Files\t*");
    switch (fc.show())
    {
    case -1:
        fl_message("%s", fc.errmsg());
        break;
    case 1:
        break;
    default:
        {
            OpenSaveData* openSaveData = (OpenSaveData*)data;
            openSaveData->filename = fc.filename();
            openSaveData->isfilename = true;
            openSaveData->drawing->open(openSaveData->filename.c_str());
        }
        break;
    }
}

static void miSaveAs_cb(Fl_Widget* /*widget*/, void* data)
{
    Fl_Native_File_Chooser fc(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
    fc.filter("Kaleidoscope files\t*.kaleido\nAll Files\t*");
    switch (fc.show())
    {
    case -1:
        fl_message("%s", fc.errmsg());
        break;
    case 1:
        break;
    default:
        {
            OpenSaveData* openSaveData = (OpenSaveData*)data;
            openSaveData->filename = fc.filename();
            if (!endswith(openSaveData->filename.c_str(), ".kaleido"))
                openSaveData->filename += ".kaleido";
            openSaveData->isfilename = true;
            openSaveData->drawing->save(openSaveData->filename.c_str());
        }
        break;
    }
}

static void miSave_cb(Fl_Widget* widget, void* data)
{
    OpenSaveData* openSaveData = (OpenSaveData*)data;
    if (!openSaveData->isfilename)
        miSaveAs_cb(widget, data);
    else
        openSaveData->drawing->save(openSaveData->filename.c_str());
}

static void miAbout_cb(Fl_Widget* /*widget*/, void* /*data*/)
{
    fl_message("%s", "Kaleidoscope 1.2\n(c) 2011 - 2021 V. Salajka\nhttp://www.franticware.com/kaleidoscope");
}

void miQuit_cb(Fl_Widget*, void*)
{
    exit(0);
}

int main(int argc, char** argv)
{
    srand(time(nullptr));
    Fl::visual(FL_RGB);
    Fl_Double_Window window(300,435+25);
    window.label("Kaleidoscope");

    Drawing drawing(10,10+25,280,280);

    static OpenSaveData openSaveData(&drawing);
    static ExportData exportData;
    Fl_Menu_Item menutable[] = {
        {"&File",0,0,0,FL_SUBMENU,0,0,0,0},
            {"&Open...", FL_CTRL+'o', miOpen_cb, &openSaveData, 0,0,0,0,0},
            {"&Save",	FL_CTRL+'s', miSave_cb, &openSaveData, 0,0,0,0,0},
            {"Save &As...",	FL_CTRL+FL_SHIFT+'s', miSaveAs_cb, &openSaveData, 0,0,0,0,0},
            {"Export TGA...",	FL_CTRL+'e', miExport_cb, &exportData, 0,0,0,0,0},
            {"&Quit",	FL_CTRL+'q', miQuit_cb, 0, 0,0,0,0,0},
            {0,0,0,0,0,0,0,0,0},
        {"&Help",0,0,0,FL_SUBMENU,0,0,0,0},
            {"&About",	0, miAbout_cb, 0, 0,0,0,0,0},
            {0,0,0,0,0,0,0,0,0},
        {0,0,0,0,0,0,0,0,0}
    };
    Fl_Menu_Bar menubar(0,0,300,25);
    menubar.menu(menutable);
    menubar.textsize(12);

    exportData.drawing = &drawing;

    // slidery
    Fl_Slider slAngle(50, 300+25, window.w()-60, 25, "Angle");
    slAngle.type(FL_HORIZONTAL);
    slAngle.align(FL_ALIGN_LEFT);
    slAngle.callback(slAngle_cb, &drawing);
    slAngle.value(INITIAL_ANGLE);
    slAngle.step(1);
    slAngle.range(0,90);
    slAngle.labelsize(12);
    drawing.m_sliders[0] = &slAngle;

    Fl_Slider slZoom(50, 325+25, window.w()-60, 25, "Zoom");
    slZoom.type(FL_HORIZONTAL);
    slZoom.align(FL_ALIGN_LEFT);
    slZoom.callback(slZoom_cb, &drawing);
    slZoom.value(INITIAL_ZOOM);
    slZoom.step(1);
    slZoom.range(-25,20);
    slZoom.labelsize(12);
    drawing.m_sliders[1] = &slZoom;

    Fl_Slider slCount(50, 350+25, window.w()-60, 25, "Count");
    slCount.type(FL_HORIZONTAL);
    slCount.align(FL_ALIGN_LEFT);
    slCount.callback(slCount_cb, &drawing);
    slCount.value(INITIAL_COUNT);
    slCount.step(1);
    slCount.range(1,200);
    slCount.labelsize(12);
    drawing.m_sliders[2] = &slCount;

    Fl_Choice chColors(50, 375+25, 150, 25, "Colors");
    const int colors_size = 5;
    static ColorsSelData colors[colors_size];
    for (int i = 0; i != colors_size; ++i)
    {
        colors[i].drawing = &drawing;
        colors[i].val = i;
    }
    chColors.add("rainbow", 0, chColors_cb, colors+0, 0);
    chColors.add("lightened rainbow", 0, chColors_cb, colors+1, 0);
    chColors.add("darkened rainbow", 0, chColors_cb, colors+2, 0);
    chColors.add("grayscale", 0, chColors_cb, colors+3, 0);
    chColors.add("full", 0, chColors_cb, colors+4, 0);
    chColors.labelsize(12);
    chColors.textsize(12);
    chColors.value(0);

    Fl_Value_Input viSize(135, 400+25, 65, 25, "Exported image size");
    viSize.value(1200);
    viSize.textsize(12);
    viSize.labelsize(12);
    exportData.size = &viSize;

    window.resizable(&drawing);
    window.end();
    window.show(argc,argv);
    return Fl::run();
}
